import 'dart:convert';
import 'dart:async';

import 'package:flutter/material.dart';
import 'package:sprintf/sprintf.dart';
import 'package:firebase_remote_config/firebase_remote_config.dart';
import 'package:telereso/telereso.dart';

import 'src/constants.dart';

/// Basic implementation for Remote Localizations , extend this class in your own Localizations , or use [telereso_generator](https://pub.dev/packages/telereso_generator) and it will be used internally
class BasicRemoteLocalizations {
  RemoteConfig _remoteConfig;
  final Locale locale;
  String localeName;
  Map<String, Map<String, dynamic>> _stringsMap = Map();

  BasicRemoteLocalizations({this.localeName, this.locale}) {
    if (locale != null) {
      this.localeName = locale.languageCode;
    }
    Telereso.instance.addRemoteChangeListener(() {
      load();
    });
  }

  // Helper method to keep the code in the widgets concise
  // Localizations are accessed using an InheritedWidget "of" syntax
  static BasicRemoteLocalizations of(BuildContext context) {
    return Localizations.of<BasicRemoteLocalizations>(
        context, BasicRemoteLocalizations);
  }

  Future<bool> load() async {
    _remoteConfig = await Telereso.instance.remoteConfig;

    final localId = _getStringsId(localeName);

    if (_remoteConfig == null) {
      _logStrings("RemoteConfig was null");
      _stringsMap[STRINGS] = {};
      _stringsMap[localId] = {};
    }

    // Load the language JSON file from the "lang" folder
    String defaultJson = _remoteConfig.getString(STRINGS);
    if (defaultJson.isEmpty) {
      _logStrings("$TAG_STRINGS: default $STRINGS was empty");
      defaultJson = "{}";
    } else {
      _logStrings("Default $STRINGS initialized");
    }

    _stringsMap[STRINGS] = json.decode(defaultJson);

    String localJson = _remoteConfig.getString(localId);

    if (localJson.isEmpty) {
      _logStrings("local $localId was empty");
      localJson = "{}";
    } else {
      _logStrings("local $localId initialized");
    }
    _stringsMap[localId] = json.decode(localJson);

    return true;
  }

  String getRemoteValue(String key) {
    final localId = _getStringsId(localeName);
    var value = _stringsMap[localId][key];

    if (value == null || value.isEmpty) {
      _logStrings("Not found in $localId");
      //   recordException(
      //       Exception("no translation for $key in ${locale.languageCode}"));
      value = _stringsMap[STRINGS][key];
    }
    if (value == null || value.isEmpty) {
      _logStrings("Not found in $STRINGS");
      // recordException(
      //     Exception("no translation for $key at defualt $defaultLocal}"));
      return null;
    }

    return value;
  }

  String translate(String key, [List<String> args]) {
    _logStrings("********************** $key **********************");
    var value = getRemoteValue(key);

    if (value == null || value.isEmpty) {
      return null;
    }

    _logStrings("value:$value");
    _logStrings("***********************************************************");
    if (args == null) {
      return value;
    }
    return sprintf(value, args);
  }

  /// Translate your key and if not found will fall back to default provided value
  /// args are a list of filling used with sprintf
  ///
  /// ```dart
  /// translateOrDefault("welcome_header","hi, welcome back %s",args: ["User"]))
  /// ```
  String translateOrDefault(String key, String def, {List<String> args}) {
    _logStrings("********************** $key **********************");
    var value = getRemoteValue(key);

    if (value == null || value.isEmpty) {
      value = def;
    }

    _logStrings("default:$def value:$value ");
    _logStrings("***********************************************************");
    if (args == null) {
      return value;
    }
    return sprintf(value, args);
  }

  String _getStringsId(String locale) {
    return "${STRINGS}_${locale}";
  }

  String _getDrawableId(String locale) {
    return "${DRAWABLE}_${locale}";
  }
}

class BasicRemoteLocalizationsDelegate
    extends LocalizationsDelegate<BasicRemoteLocalizations> {
  const BasicRemoteLocalizationsDelegate();

  @override
  bool isSupported(Locale locale) =>
      Telereso.instance.supportedLocals.contains(locale.languageCode);

  @override
  Future<BasicRemoteLocalizations> load(Locale locale) async {
    var localizations = BasicRemoteLocalizations(locale: locale);
    await localizations.load();
    return localizations;
  }

  @override
  bool shouldReload(BasicRemoteLocalizationsDelegate old) => false;
}

extension DefaultMap<K, V> on Map<K, V> {
  V getOrElse(K key, V defaultValue) {
    if (this.containsKey(key)) {
      return this[key];
    } else {
      return defaultValue;
    }
  }
}


_logStrings(String log) {
  if (Telereso.instance.isStringLogEnabled) {
    print("$TAG_STRINGS $log");
  }
}